package it.openutils.spring.rmibernate.server.managers;

import it.openutils.spring.rmibernate.shared.managers.HibernateLazyService;

import java.io.Serializable;
import java.rmi.RemoteException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.lang.StringUtils;
import org.hibernate.Hibernate;
import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.collection.PersistentCollection;
import org.hibernate.collection.PersistentSortedMap;
import org.hibernate.collection.PersistentSortedSet;
import org.hibernate.proxy.HibernateProxy;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.orm.hibernate3.HibernateCallback;
import org.springframework.orm.hibernate3.HibernateTemplate;


/**
 * Server side implementation of remote lazy loading calls
 * @author mmolaschi
 * @version $Id: HibernateLazyServiceImpl.java 482 2007-10-26 08:56:04Z fcarone $
 */
public class HibernateLazyServiceImpl implements HibernateLazyService
{

    /**
     * hibernate sessionfactory
     */
    private SessionFactory sessionFactory;

    /**
     * Logger.
     */
    private static Logger log = LoggerFactory.getLogger(HibernateLazyServiceImpl.class);

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public Object invoke(final String className, final String methodName, final Serializable id) throws RemoteException
    {
        log.debug("Initializing lazy property.");
        return invoke(className, methodName, id, true);
    }

    private Object invoke(final String className, final String methodName, final Serializable id, final boolean init)
        throws RemoteException
    {
        try
        {
            HibernateTemplate ht = new HibernateTemplate(sessionFactory);
            return ht.execute(new HibernateCallback()
            {
                public Object doInHibernate(Session session) throws HibernateException, SQLException
                {
                    try
                    {
                        log.debug("Loading class: {}, id {}, method {}", new Object[] {className, id, methodName});
                        Object entity = sessionFactory.getCurrentSession().load(Class.forName(className), id);
                        Object retVal = PropertyUtils.getProperty(entity, methodName);
                        if (init)
                        {
                            return init(retVal);
                        }
                        return retVal;
                    }
                    catch (Exception ex)
                    {
                        log.error("{}", ex);
                        throw new HibernateException(ex);
                    }
                }

            });
        }
        catch (HibernateException ex)
        {
            throw new RemoteException(ex.getMessage(), ex);
        }
    }

    @SuppressWarnings("unchecked")
    private Object init(Object o)
    {
        if (o == null)
        {
            return null;
        }

        // get class
        if (o instanceof HibernateProxy)
        {
            Hibernate.initialize(o);
            return o;
        }

        if (o instanceof List)
        {
            return new ArrayList((List) o);
        }

        if (o instanceof SortedSet)
        {
            PersistentSortedSet pss = (PersistentSortedSet) o;
            SortedSet ss = new TreeSet(pss.comparator());
            ss.addAll((Collection) o);
            return ss;
        }

        if (o instanceof Set)
        {
            return new HashSet((Collection) o);
        }

        if (o instanceof SortedMap)
        {
            PersistentSortedMap psm = (PersistentSortedMap) o;
            SortedMap sm = new TreeMap(psm.comparator());
            sm.putAll((Map) o);
            return sm;
        }

        if (o instanceof Map)
        {
            return new HashMap((Map) o);
        }

        // it should never reach this line...
        return null;
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public Object invoke(final String className, final String propertyName, final Serializable id, final int from,
        final int size) throws RemoteException
    {
        try
        {
            // get lazy prop
            final Object lazy = invoke(className, propertyName, id, false);

            if (lazy == null)
            {
                return null;
            }

            // get class
            Class fClass = lazy.getClass();

            // if is a single object proxy
            if (lazy instanceof HibernateProxy)
            {
                return init(lazy);
            }
            else
            {
                // it's a list

                // calculate list size
                long count = this.count(className, propertyName, id);

                // no row at desired rownum
                if (from > count)
                {
                    return null;
                }

                // the list is already right and can be processed
                if (from <= 0 && size >= count)
                {
                    return init(lazy);
                }
            }

            // filter collection
            HibernateTemplate ht = new HibernateTemplate(sessionFactory);
            List list = ht.executeFind(new HibernateCallback()
            {

                public Object doInHibernate(Session session) throws HibernateException, SQLException
                {
                    Query q = session.createFilter(lazy, "");
                    q.setMaxResults(size);
                    q.setFirstResult(from);

                    return q.list();
                }

            });

            // process list
            if (list == null || list.size() == 0)
            {
                return null;
            }

            if (lazy instanceof PersistentCollection)
            {
                if (lazy instanceof List)
                {
                    return list;
                }
                if (lazy instanceof SortedSet)
                {
                    PersistentSortedSet pss = (PersistentSortedSet) lazy;
                    SortedSet ss = new TreeSet(pss.comparator());
                    ss.addAll(list);
                    return ss;
                }
                if (lazy instanceof Set)
                {
                    return new HashSet(list);
                }
                if (lazy instanceof SortedMap)
                {
                    PersistentSortedMap psm = (PersistentSortedMap) invoke(className, propertyName, id);
                    SortedMap sm = new TreeMap(psm.comparator());
                    for (Object o : list)
                    {
                        Object[] array = (Object[]) o;
                        sm.put(array[0], array[1]);
                    }
                    return sm;
                }
                if (lazy instanceof Map)
                {
                    Map m = new HashMap();
                    for (Object o : list)
                    {
                        Object[] array = (Object[]) o;
                        m.put(array[0], array[1]);
                    }
                    return m;
                }
            }

            // it should never reach this point ?!?!
            return null;
        }
        catch (Exception ex)
        {
            throw new RemoteException("Error lazy loading", ex);
        }
    }

    /**
     * {@inheritDoc}
     */
    public long count(final String className, final String propertyName, final Serializable id) throws RemoteException
    {
        HibernateTemplate ht = new HibernateTemplate(sessionFactory);
        return ((Number) ht.execute(new HibernateCallback()
        {

            public Object doInHibernate(Session session) throws HibernateException, SQLException
            {
                StringBuilder query = getBaseQuery(className, propertyName, true);
                Query q = session.createQuery(query.toString());
                q.setParameter(0, id);

                return q.uniqueResult();
            }

        })).longValue();
    }

    private StringBuilder getBaseQuery(String className, String propertyName, boolean count)
    {
        String entityName = StringUtils.substringAfterLast(className, ".");

        String propertyId = sessionFactory.getClassMetadata(className).getIdentifierPropertyName();

        StringBuilder query = new StringBuilder("select ");
        if (count)
        {
            query.append("count(");
        }
        query.append("el");
        if (count)
        {
            query.append(")");
        }
        query
            .append(" from ")
            .append(entityName)
            .append(" cl inner join cl.")
            .append(propertyName)
            .append(" el ")
            .append(" where cl.")
            .append(propertyId)
            .append(" = ?");

        return query;
    }

    /**
     * Sets the factory.
     * @param factory the factory to set
     */
    public void setSessionFactory(SessionFactory factory)
    {
        this.sessionFactory = factory;
    }

}
